import argparse
import sys
import csv
from itertools import islice
from .parse_types import *


def create():
    """Set up and return the parser"""
    script_name = "csvprint"
    parser = argparse.ArgumentParser(
        description="Command line utility for pretty printing csv files.",
        formatter_class=argparse.RawTextHelpFormatter,
        prog=script_name,
    )
    parser.add_argument("filename", type=str, help="file to pretty print", nargs="?")
    parser.add_argument(
        "-s",
        "--separator",
        default=",",
        help="separator/delimiter used in csv file\ndefault is comma: ','\nuse 'tab' for tab separated files\n",
        type=separator,
    )
    parser.add_argument(
        "-n",
        "--rows",
        type=positive_integer,
        default=sys.maxsize,
        help="number of rows to show",
    )
    parser.add_argument(
        "-a",
        "--align",
        nargs="+",
        default=["<"],
        help="which alignment to use\ndefault is left\nchoices: {l, r}\n"
        + "can provide a list, in which case\none choice for each column",
        type=alignment,
    )
    parser.add_argument(
        "-d",
        "--decorator",
        type=str,
        default="",
        help="which string/decorator to use in spacing",
    )
    parser.add_argument(
        "-p", "--padding", type=non_negative_integer, default=1, help="padding"
    )
    parser.add_argument(
        "-c",
        "--columns",
        type=int,
        nargs="*",
        metavar="column",
        help="select columns to print (1-indexed)",
    )
    parser.add_argument(
        "--numeric",
        type=decimal_number_formatting,
        nargs="*",
        help="format numerical columns",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--header", action="store_true", help="header decoration")
    group.add_argument("--markdown", action="store_true", help="output markdown table")
    group.add_argument("--latex", action="store_true", help="output latex table")
    return parser


def print_message_and_exit(parser, message):
    """Print parser usage and the error message provided"""
    script_name = "csvprint"
    parser.print_usage()
    print("{}: error:".format(script_name), end=" ")
    print(message)
    sys.exit()


def file_error_checking(parser, filename):
    """Check the file or pipe provided to the parser is correct"""
    reading_from_pipe = not sys.stdin.isatty() and filename == None
    if reading_from_pipe:
        file = sys.stdin
    elif filename == None:
        print_message_and_exit(parser, "required: filename or pipe")
    else:
        try:
            file = open(filename)
        except FileNotFoundError:
            print_message_and_exit(parser, "no such file: {}".format(filename))
    return file


def parse_cli_arguments(parser):
    """Parse command line arguments and return dictionary with configuration to be used"""
    args = vars(parser.parse_args())
    args["csvfile"] = file_error_checking(parser, args["filename"])
    if args["markdown"]:
        args["decorator"] = "|"
    if args["latex"]:
        args["decorator"] = "&"
    return args


def lstrip_cells_in_row(row):
    return [cell.lstrip() for cell in row]


def list_of_numbers_is_sorted_and_contains_unique_elements(numbers):
    return sorted(list(set(numbers))) == numbers


def store_content(parser, args):
    """Store content in file and extract relevant configurations to the dictionary"""
    csvreader = csv.reader(args["csvfile"], delimiter=args["separator"])
    header = next(csvreader)
    args["num_columns"] = len(header)
    args["content"] = [lstrip_cells_in_row(header)]
    row_number = 0
    for row_number, row in enumerate(islice(csvreader, args["rows"] - 1)):
        args["num_columns"] = max(len(row), args["num_columns"])
        args["content"].append(lstrip_cells_in_row(row))
    args["rows"] = row_number + 1
    align_all_columns_equally = len(args["align"]) == 1
    length_of_alignment_and_columns_differ = len(args["align"]) != args["num_columns"]
    if align_all_columns_equally:
        args["align"] = [args["align"][0]] * args["num_columns"]
    elif length_of_alignment_and_columns_differ:
        print_message_and_exit(
            parser, "argument -a/--align: only one argument or one per column"
        )
    if args["columns"] is None:
        args["columns"] = list(range(1, args["num_columns"] + 1))
    elif not all(0 < c <= args["num_columns"] for c in args["columns"]):
        print_message_and_exit(
            parser, "argument -c/--columns: column number out of range"
        )
    elif not list_of_numbers_is_sorted_and_contains_unique_elements(args["columns"]):
        print_message_and_exit(
            parser,
            "argument -c/--columns: column numbers must be unique and in ascending order",
        )



def copy_nested_list(l):
    """Return a copy of list l to one level of nesting"""
    return [list(i) for i in l]


def normalize_table(table, n):
    """Return a normalized version of table such that it has n columns in each row, possibly with empty cells"""
    normalized_table = copy_nested_list(table)
    for row in normalized_table:
        while len(row) < n:
            row.append("")
    return normalized_table


def column_widths(table):
    """Get the maximum size for each column in table"""
    return [max(map(len, col)) for col in zip(*table)]


def align_table(table, align):
    """Return table justified according to align"""
    widths = column_widths(table)
    new_table = copy_nested_list(table)
    for row in new_table:
        for cell_num, cell in enumerate(row):
            row[cell_num] = "{:{align}{width}}".format(
                cell, align=align[cell_num], width=widths[cell_num]
            )
    return new_table


def header_line(length, border="-"):
    """Return header of length"""
    return border * length


def md_alignment(alignment):
    """Given alignment option, return corresponding markdown"""
    if alignment == "<":
        return ":", "-"
    elif alignment == ">":
        return "-", ":"
    else:
        return "-", "-"


def latex_alignment(alignment):
    """Given alignment option, return corresponding latex"""
    if alignment == "<":
        return "l"
    elif alignment == ">":
        return "r"


def markdown_header(table, align):
    """Get markdown header for table according to given alignment options"""
    widths = column_widths(table)
    header_line = []
    for i, w in enumerate(widths):
        if w < 3:
            raise ValueError
        left, right = md_alignment(align[i])
        header_line.append(left + "-" * (w - 2) + right)
    return header_line


def add_markdown_header(table, align):
    """Add markdown header lines to table"""
    new_table = copy_nested_list(table)
    new_table.insert(1, markdown_header(table, align))
    return new_table


def add_header(table, align):
    """Add header lines to table"""
    widths = column_widths(table)
    new_table = copy_nested_list(table)
    for index in (0, 2):
        new_table.insert(index, [header_line(w) for w in widths])
    return new_table


def add_padding(table, padding):
    """Return a version of table which is padded according to inputted padding"""
    new_table = copy_nested_list(table)
    for i, row in enumerate(new_table):
        padding_string = " " * padding
        for j, cell in enumerate(row):
            left = padding_string
            right = padding_string
            if j == 0:
                left = ""
            elif j == len(row) - 1:
                right = ""
            new_table[i][j] = left + new_table[i][j] + right
    return new_table


def join_columns_with_divider(table, decorator):
    """Join each line in table with the decorator string between each cell"""
    return [decorator.join(row) for row in table]


def join_formatted_lines(lines):
    """Return the finished output"""
    return "\n".join(lines)


def add_latex_line_endings(lines):
    """Add latex newline character to each line"""
    return [line + r" \\" for line in lines]


def add_latex_table_environment(lines, number_of_columns, alignment):
    """Add latex environment specification"""
    lines = list(lines)
    latex_alignments = [latex_alignment(j) for j in alignment]
    alignment_line = "{{{}}}".format("".join(latex_alignments))
    lines.insert(0, r"\begin{tabular}" + alignment_line)
    lines.append(r"\end{tabular}")
    return lines


def right_strip_lines(lines):
    """Remove trailing spaces on each line"""
    return [line.rstrip() for line in lines]


def select_columns_from_table(table, columns, alignment):
    new_table = []
    new_alignment = []
    for row in table:
        new_row = []
        for column_number in columns:
            index_number = column_number - 1
            cell = row[index_number]
            new_row.append(cell)
        new_table.append(new_row)
    for column_number in columns:
        index_number = column_number - 1
        new_alignment.append(alignment[index_number])
    return new_table, new_alignment


def run_pipeline(args):
    """Run the printing pipeline according to arguments given"""
    raw_table = args["content"]
    alignment = args["align"]
    padding = args["padding"]
    output_as_markdown = args["markdown"]
    output_as_latex = args["latex"]
    output_as_header = args["header"]
    decorator = args["decorator"]
    columns_to_print = args["columns"]
    number_of_columns = len(columns_to_print)
    if args["numeric"] is not None:
        transposed_table = list(zip(*raw_table))
        for column_number, decimal_numbers in args["numeric"]:
            numeric_column = transposed_table[column_number - 1]
            new_column = []
            for number in numeric_column:
                try:
                    number = f"%.{decimal_numbers}f" % float(number)
                except:
                    number = number
                new_column.append(number)
            transposed_table[column_number - 1] = new_column
        raw_table = list(zip(*transposed_table))
    raw_table_subset, alignment_subset = select_columns_from_table(
        raw_table, columns_to_print, alignment
    )
    normalized_table = normalize_table(raw_table_subset, number_of_columns)
    justified_table = align_table(normalized_table, alignment_subset)
    padded_table = add_padding(justified_table, padding)
    if output_as_markdown:
        padded_table = add_markdown_header(padded_table, alignment_subset)
    if output_as_header:
        padded_table = add_header(padded_table, alignment_subset)
    lines = join_columns_with_divider(padded_table, decorator)
    if output_as_latex:
        lines = add_latex_line_endings(lines)
        lines = add_latex_table_environment(lines, number_of_columns, alignment)
    else:
        lines = right_strip_lines(lines)
    finished_output = join_formatted_lines(lines)
    return finished_output



import argparse


def alignment(direction):
    """Parse the direction, possibly raising error"""
    if direction == "l":
        return "<"
    elif direction == "r":
        return ">"
    else:
        raise argparse.ArgumentTypeError("must be l for left or r for right")


def separator(string):
    """Parse the separator, possibly raising error"""
    options = {"tab": "\t", "comma": ","}
    s = options.get(string, string)
    if len(s) == 1:
        return s
    else:
        raise argparse.ArgumentTypeError("must be a 1-character string")


def positive_integer(n):
    """Parse positive integer, possibly raising error"""
    message = "must be a positive integer"
    try:
        n = int(n)
    except ValueError:
        raise argparse.ArgumentTypeError(message)
    if n > 0:
        return n
    else:
        raise argparse.ArgumentTypeError(message)


def non_negative_integer(n):
    """Parse non-negative integer, possibly raising error"""
    message = "must be a non-negative integer"
    try:
        n = int(n)
    except ValueError:
        raise argparse.ArgumentTypeError(message)
    if n >= 0:
        return n
    else:
        raise argparse.ArgumentTypeError(message)


def decimal_number_formatting(arg):
    """On format [column_number]:[decimal_numbers]"""
    split = arg.split(":")
    if len(split) != 2:
        raise argparse.ArgumentTypeError(
            "must be on format [column_number]:[decimal_numbers]"
        )
    column, decimals = split
    try:
        column = int(column)
    except:
        raise argparse.ArgumentTypeError("column must be integer")
    if column < 1:
        raise argparse.ArgumentTypeError("column must be positive")
    try:
        decimals = int(decimals)
    except:
        raise argparse.ArgumentTypeError("decimals must be integer")
    if column < 0:
        raise argparse.ArgumentTypeError("decimals must be non-negative")
    return column, decimals



from setuptools import setup

with open("README.md", "r") as readme:
    LONG_DESCRIPTION = readme.read()

setup(
    name="csvprint",
    version="0.6.1",
    description="Print csv files in columnated format, either plain or as a Markdown or LaTeX table",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=["csvprint"],
    license="MIT",
    scripts=["bin/csvprint"],
    author="Vegard Stikbakke",
    author_email="vegard.stikbakke@gmail.com",
    url="https://github.com/vegarsti/csvprint",
    keywords="csv",
)
